@@ -10,7 +10,10 @@ Notes:
 
 # External modules
 import requests
+import asyncio
+import warnings
 
+from aiohttp import ClientSession, TCPConnector
 from pkg_resources import get_distribution, DistributionNotFound
 
 # Obtain package version for the user-agent. Uses the installed version by
@@ -96,7 +99,7 @@ class Epidata:
       raise Exception('`issues` and `lag` are mutually exclusive')
     # Set up request
     params = {
-      'source': 'fluview',
+      'endpoint': 'fluview',
       'regions': Epidata._list(regions),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -115,7 +118,7 @@ class Epidata:
     """Fetch FluView metadata."""
     # Set up request
     params = {
-      'source': 'fluview_meta',
+      'endpoint': 'fluview_meta',
     }
     # Make the API call
     return Epidata._request(params)
@@ -131,7 +134,7 @@ class Epidata:
       raise Exception('`issues` and `lag` are mutually exclusive')
     # Set up request
     params = {
-      'source': 'fluview_clinical',
+      'endpoint': 'fluview_clinical',
       'regions': Epidata._list(regions),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -153,7 +156,7 @@ class Epidata:
       raise Exception('`issues` and `lag` are mutually exclusive')
     # Set up request
     params = {
-      'source': 'flusurv',
+      'endpoint': 'flusurv',
       'locations': Epidata._list(locations),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -175,7 +178,7 @@ class Epidata:
       raise Exception('`issues` and `lag` are mutually exclusive')
     # Set up request
     params = {
-      'source': 'paho_dengue',
+      'endpoint': 'paho_dengue',
       'regions': Epidata._list(regions),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -197,7 +200,7 @@ class Epidata:
       raise Exception('`issues` and `lag` are mutually exclusive')
     # Set up request
     params = {
-      'source': 'ecdc_ili',
+      'endpoint': 'ecdc_ili',
       'regions': Epidata._list(regions),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -219,7 +222,7 @@ class Epidata:
       raise Exception('`issues` and `lag` are mutually exclusive')
     # Set up request
     params = {
-      'source': 'kcdc_ili',
+      'endpoint': 'kcdc_ili',
       'regions': Epidata._list(regions),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -239,7 +242,7 @@ class Epidata:
       raise Exception('`locations` and `epiweeks` are both required')
     # Set up request
     params = {
-      'source': 'gft',
+      'endpoint': 'gft',
       'locations': Epidata._list(locations),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -255,7 +258,7 @@ class Epidata:
       raise Exception('`auth`, `locations`, `epiweeks`, and `query` are all required')
     # Set up request
     params = {
-      'source': 'ght',
+      'endpoint': 'ght',
       'auth': auth,
       'locations': Epidata._list(locations),
       'epiweeks': Epidata._list(epiweeks),
@@ -275,7 +278,7 @@ class Epidata:
       raise Exception('exactly one of `dates` and `epiweeks` is required')
     # Set up request
     params = {
-      'source': 'twitter',
+      'endpoint': 'twitter',
       'auth': auth,
       'locations': Epidata._list(locations),
     }
@@ -297,7 +300,7 @@ class Epidata:
       raise Exception('exactly one of `dates` and `epiweeks` is required')
     # Set up request
     params = {
-      'source': 'wiki',
+      'endpoint': 'wiki',
       'articles': Epidata._list(articles),
       'language': language,
     }
@@ -319,7 +322,7 @@ class Epidata:
       raise Exception('`auth`, `epiweeks`, and `locations` are all required')
     # Set up request
     params = {
-      'source': 'cdc',
+      'endpoint': 'cdc',
       'auth': auth,
       'epiweeks': Epidata._list(epiweeks),
       'locations': Epidata._list(locations),
@@ -336,7 +339,7 @@ class Epidata:
       raise Exception('`auth`, `epiweeks`, and `locations` are all required')
     # Set up request
     params = {
-      'source': 'quidel',
+      'endpoint': 'quidel',
       'auth': auth,
       'epiweeks': Epidata._list(epiweeks),
       'locations': Epidata._list(locations),
@@ -353,7 +356,7 @@ class Epidata:
       raise Exception('`auth`, `location`, and `epiweeks` are all required')
     # Set up request
     params = {
-      'source': 'norostat',
+      'endpoint': 'norostat',
       'auth': auth,
       'location': location,
       'epiweeks': Epidata._list(epiweeks),
@@ -370,7 +373,7 @@ class Epidata:
       raise Exception('`auth` is required')
     # Set up request
     params = {
-      'source': 'meta_norostat',
+      'endpoint': 'meta_norostat',
       'auth': auth,
     }
     # Make the API call
@@ -408,7 +411,7 @@ class Epidata:
 
     # Set up request
     params = {
-      'source': 'afhsb',
+      'endpoint': 'afhsb',
       'auth': auth,
       'locations': Epidata._list(locations),
       'epiweeks': Epidata._list(epiweeks),
@@ -426,7 +429,7 @@ class Epidata:
       raise Exception('`auth` is required')
     # Set up request
     params = {
-      'source': 'meta_afhsb',
+      'endpoint': 'meta_afhsb',
       'auth': auth,
     }
     # Make the API call
@@ -443,7 +446,7 @@ class Epidata:
       raise Exception('`issues` and `lag` are mutually exclusive')
     # Set up request
     params = {
-      'source': 'nidss_flu',
+      'endpoint': 'nidss_flu',
       'regions': Epidata._list(regions),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -463,7 +466,7 @@ class Epidata:
       raise Exception('`locations` and `epiweeks` are both required')
     # Set up request
     params = {
-      'source': 'nidss_dengue',
+      'endpoint': 'nidss_dengue',
       'locations': Epidata._list(locations),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -479,7 +482,7 @@ class Epidata:
       raise Exception('`system` and `epiweek` are both required')
     # Set up request
     params = {
-      'source': 'delphi',
+      'endpoint': 'delphi',
       'system': system,
       'epiweek': epiweek,
     }
@@ -495,7 +498,7 @@ class Epidata:
       raise Exception('`auth`, `names`, `locations`, and `epiweeks` are all required')
     # Set up request
     params = {
-      'source': 'sensors',
+      'endpoint': 'sensors',
       'auth': auth,
       'names': Epidata._list(names),
       'locations': Epidata._list(locations),
@@ -513,7 +516,7 @@ class Epidata:
       raise Exception('`auth`, `names`, `locations`, and `epiweeks` are all required')
     # Set up request
     params = {
-      'source': 'dengue_sensors',
+      'endpoint': 'dengue_sensors',
       'auth': auth,
       'names': Epidata._list(names),
       'locations': Epidata._list(locations),
@@ -531,7 +534,7 @@ class Epidata:
       raise Exception('`locations` and `epiweeks` are both required')
     # Set up request
     params = {
-      'source': 'nowcast',
+      'endpoint': 'nowcast',
       'locations': Epidata._list(locations),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -547,7 +550,7 @@ class Epidata:
       raise Exception('`locations` and `epiweeks` are both required')
     # Set up request
     params = {
-      'source': 'dengue_nowcast',
+      'endpoint': 'dengue_nowcast',
       'locations': Epidata._list(locations),
       'epiweeks': Epidata._list(epiweeks),
     }
@@ -558,7 +561,7 @@ class Epidata:
   @staticmethod
   def meta():
     """Fetch API metadata."""
-    return Epidata._request({'source': 'meta'})
+    return Epidata._request({'endpoint': 'meta'})
 
   # Fetch Delphi's COVID-19 Surveillance Streams
   @staticmethod
@@ -576,7 +579,7 @@ class Epidata:
       raise Exception('`issues` and `lag` are mutually exclusive')
     # Set up request
     params = {
-      'source': 'covidcast',
+      'endpoint': 'covidcast',
       'data_source': data_source,
       'signals': Epidata._list(signals),
       'time_type': time_type,
@@ -605,7 +608,7 @@ class Epidata:
   @staticmethod
   def covidcast_meta():
     """Fetch Delphi's COVID-19 Surveillance Streams metadata"""
-    return Epidata._request({'source': 'covidcast_meta'})
+    return Epidata._request({'endpoint': 'covidcast_meta'})
 
   # Fetch COVID hospitalization data
   @staticmethod
@@ -616,7 +619,7 @@ class Epidata:
       raise Exception('`states` and `dates` are both required')
     # Set up request
     params = {
-      'source': 'covid_hosp',
+      'endpoint': 'covid_hosp',
       'states': Epidata._list(states),
       'dates': Epidata._list(dates),
     }
@@ -665,3 +668,67 @@ class Epidata:
       raise Exception('one of `state`, `ccn`, `city`, `zip`, or `fips_code` is required')
     # Make the API call
     return Epidata._request(params)
+
+  # Fetch Delphi's COVID-19 Nowcast sensors
+  @staticmethod
+  def covidcast_nowcast(
+          data_source, signals, sensor_names, time_type, geo_type,
+          time_values, geo_value, as_of=None, issues=None, lag=None, **kwargs):
+    """Fetch Delphi's COVID-19 Nowcast sensors"""
+    # Check parameters
+    if data_source is None or signals is None or time_type is None or geo_type is None or time_values is None or geo_value is None or sensor_names is None:
+      raise Exception('`data_source`, `signals`, `sensor_names`, `time_type`, `geo_type`, `time_values`, and `geo_value` are all required')
+    if issues is not None and lag is not None:
+      raise Exception('`issues` and `lag` are mutually exclusive')
+    # Set up request
+    params = {
+      'source': 'covidcast_nowcast',
+      'data_source': data_source,
+      'signals': Epidata._list(signals),
+      'sensor_names': Epidata._list(sensor_names),
+      'time_type': time_type,
+      'geo_type': geo_type,
+      'time_values': Epidata._list(time_values)
+    }
+
+    if isinstance(geo_value, (list, tuple)):
+      params['geo_values'] = ','.join(geo_value)
+    else:
+      params['geo_value'] = geo_value
+    if as_of is not None:
+      params['as_of'] = as_of
+    if issues is not None:
+      params['issues'] = Epidata._list(issues)
+    if lag is not None:
+      params['lag'] = lag
+
+    if 'format' in kwargs:
+      params['format'] = kwargs['format']
+
+    # Make the API call
+    return Epidata._request(params)
+
+  @staticmethod
+  def async_epidata(param_list, batch_size=50):
+    """Make asynchronous Epidata calls for a list of parameters."""
+    async def async_get(params, session):
+      """Helper function to make Epidata GET requests."""
+      async with session.get(Epidata.BASE_URL, params=params) as response:
+        response.raise_for_status()
+        return await response.json(), params
+
+    async def async_make_calls(param_combos):
+      """Helper function to asynchronously make and aggregate Epidata GET requests."""
+      tasks = []
+      connector = TCPConnector(limit=batch_size)
+      async with ClientSession(connector=connector) as session:
+        for param in param_combos:
+          task = asyncio.ensure_future(async_get(param, session))
+          tasks.append(task)
+        responses = await asyncio.gather(*tasks)
+        return responses
+
+    loop = asyncio.get_event_loop()
+    future = asyncio.ensure_future(async_make_calls(param_list))
+    responses = loop.run_until_complete(future)
+    return responses
